import operator
import re
import llvmlite.binding as ll
import numba
import numpy as np
from llvmlite import ir as lir
from numba.core import cgutils, types
from numba.core.typing.templates import AbstractTemplate, AttributeTemplate, bound_function, infer_getattr, infer_global, signature
from numba.extending import intrinsic, lower_cast, make_attribute_wrapper, models, overload, overload_attribute, register_jitable, register_model
from numba.parfors.array_analysis import ArrayAnalysis
import bodo
from bodo.libs import hstr_ext
from bodo.utils.typing import BodoError, get_overload_const_int, get_overload_const_str, is_overload_constant_int, is_overload_constant_str


def unliteral_all(args):
    return tuple(types.unliteral(a) for a in args)


ll.add_symbol('del_str', hstr_ext.del_str)
ll.add_symbol('unicode_to_utf8', hstr_ext.unicode_to_utf8)
ll.add_symbol('memcmp', hstr_ext.memcmp)
ll.add_symbol('int_to_hex', hstr_ext.int_to_hex)
string_type = types.unicode_type


@numba.njit
def contains_regex(e, in_str):
    with numba.objmode(res='bool_'):
        res = bool(e.search(in_str))
    return res


@numba.generated_jit
def str_findall_count(regex, in_str):

    def _str_findall_count_impl(regex, in_str):
        with numba.objmode(res='int64'):
            res = len(regex.findall(in_str))
        return res
    return _str_findall_count_impl


utf8_str_type = types.ArrayCTypes(types.Array(types.uint8, 1, 'C'))


@intrinsic
def unicode_to_utf8_and_len(typingctx, str_typ=None):
    assert str_typ in (string_type, types.Optional(string_type)) or isinstance(
        str_typ, types.StringLiteral)
    dule__zxfr = types.Tuple([utf8_str_type, types.int64])

    def codegen(context, builder, sig, args):
        ekv__jphb, = args
        tii__tqpg = cgutils.create_struct_proxy(string_type)(context,
            builder, value=ekv__jphb)
        qwk__nhnt = cgutils.create_struct_proxy(utf8_str_type)(context, builder
            )
        inae__htp = cgutils.create_struct_proxy(dule__zxfr)(context, builder)
        is_ascii = builder.icmp_unsigned('==', tii__tqpg.is_ascii, lir.
            Constant(tii__tqpg.is_ascii.type, 1))
        with builder.if_else(is_ascii) as (bukq__cuu, powl__fccbw):
            with bukq__cuu:
                context.nrt.incref(builder, string_type, ekv__jphb)
                qwk__nhnt.data = tii__tqpg.data
                qwk__nhnt.meminfo = tii__tqpg.meminfo
                inae__htp.f1 = tii__tqpg.length
            with powl__fccbw:
                vuu__vgt = lir.FunctionType(lir.IntType(64), [lir.IntType(8
                    ).as_pointer(), lir.IntType(8).as_pointer(), lir.
                    IntType(64), lir.IntType(32)])
                smo__yjw = cgutils.get_or_insert_function(builder.module,
                    vuu__vgt, name='unicode_to_utf8')
                owrl__mcy = context.get_constant_null(types.voidptr)
                kgjc__turpl = builder.call(smo__yjw, [owrl__mcy, tii__tqpg.
                    data, tii__tqpg.length, tii__tqpg.kind])
                inae__htp.f1 = kgjc__turpl
                tctnc__typqt = builder.add(kgjc__turpl, lir.Constant(lir.
                    IntType(64), 1))
                qwk__nhnt.meminfo = context.nrt.meminfo_alloc_aligned(builder,
                    size=tctnc__typqt, align=32)
                qwk__nhnt.data = context.nrt.meminfo_data(builder,
                    qwk__nhnt.meminfo)
                builder.call(smo__yjw, [qwk__nhnt.data, tii__tqpg.data,
                    tii__tqpg.length, tii__tqpg.kind])
                builder.store(lir.Constant(lir.IntType(8), 0), builder.gep(
                    qwk__nhnt.data, [kgjc__turpl]))
        inae__htp.f0 = qwk__nhnt._getvalue()
        return inae__htp._getvalue()
    return dule__zxfr(string_type), codegen


def unicode_to_utf8(s):
    return s


@overload(unicode_to_utf8)
def overload_unicode_to_utf8(s):
    return lambda s: unicode_to_utf8_and_len(s)[0]


@overload(max)
def overload_builtin_max(lhs, rhs):
    if lhs == types.unicode_type and rhs == types.unicode_type:

        def impl(lhs, rhs):
            return lhs if lhs > rhs else rhs
        return impl


@overload(min)
def overload_builtin_min(lhs, rhs):
    if lhs == types.unicode_type and rhs == types.unicode_type:

        def impl(lhs, rhs):
            return lhs if lhs < rhs else rhs
        return impl


@intrinsic
def memcmp(typingctx, dest_t, src_t, count_t=None):

    def codegen(context, builder, sig, args):
        vuu__vgt = lir.FunctionType(lir.IntType(32), [lir.IntType(8).
            as_pointer(), lir.IntType(8).as_pointer(), lir.IntType(64)])
        gllw__eiib = cgutils.get_or_insert_function(builder.module,
            vuu__vgt, name='memcmp')
        return builder.call(gllw__eiib, args)
    return types.int32(types.voidptr, types.voidptr, types.intp), codegen


def int_to_str_len(n):
    return len(str(n))


@overload(int_to_str_len)
def overload_int_to_str_len(n):
    duxn__rwsj = n(10)

    def impl(n):
        if n == 0:
            return 1
        budya__bfr = 0
        if n < 0:
            n = -n
            budya__bfr += 1
        while n > 0:
            n = n // duxn__rwsj
            budya__bfr += 1
        return budya__bfr
    return impl


class StdStringType(types.Opaque):

    def __init__(self):
        super(StdStringType, self).__init__(name='StdStringType')


std_str_type = StdStringType()
register_model(StdStringType)(models.OpaqueModel)
del_str = types.ExternalFunction('del_str', types.void(std_str_type))
get_c_str = types.ExternalFunction('get_c_str', types.voidptr(std_str_type))
dummy_use = numba.njit(lambda a: None)


@overload(int)
def int_str_overload(in_str, base=10):
    if in_str == string_type:
        if is_overload_constant_int(base) and get_overload_const_int(base
            ) == 10:

            def _str_to_int_impl(in_str, base=10):
                val = _str_to_int64(in_str._data, in_str._length)
                dummy_use(in_str)
                return val
            return _str_to_int_impl

        def _str_to_int_base_impl(in_str, base=10):
            val = _str_to_int64_base(in_str._data, in_str._length, base)
            dummy_use(in_str)
            return val
        return _str_to_int_base_impl


@infer_global(float)
class StrToFloat(AbstractTemplate):

    def generic(self, args, kws):
        assert not kws
        [gjbit__ghw] = args
        if isinstance(gjbit__ghw, StdStringType):
            return signature(types.float64, gjbit__ghw)
        if gjbit__ghw == string_type:
            return signature(types.float64, gjbit__ghw)


ll.add_symbol('init_string_const', hstr_ext.init_string_const)
ll.add_symbol('get_c_str', hstr_ext.get_c_str)
ll.add_symbol('str_to_int64', hstr_ext.str_to_int64)
ll.add_symbol('str_to_uint64', hstr_ext.str_to_uint64)
ll.add_symbol('str_to_int64_base', hstr_ext.str_to_int64_base)
ll.add_symbol('str_to_float64', hstr_ext.str_to_float64)
ll.add_symbol('str_to_float32', hstr_ext.str_to_float32)
ll.add_symbol('get_str_len', hstr_ext.get_str_len)
ll.add_symbol('str_from_float32', hstr_ext.str_from_float32)
ll.add_symbol('str_from_float64', hstr_ext.str_from_float64)
get_std_str_len = types.ExternalFunction('get_str_len', signature(types.
    intp, std_str_type))
init_string_from_chars = types.ExternalFunction('init_string_const',
    std_str_type(types.voidptr, types.intp))
_str_to_int64 = types.ExternalFunction('str_to_int64', signature(types.
    int64, types.voidptr, types.int64))
_str_to_uint64 = types.ExternalFunction('str_to_uint64', signature(types.
    uint64, types.voidptr, types.int64))
_str_to_int64_base = types.ExternalFunction('str_to_int64_base', signature(
    types.int64, types.voidptr, types.int64, types.int64))


def gen_unicode_to_std_str(context, builder, unicode_val):
    tii__tqpg = cgutils.create_struct_proxy(string_type)(context, builder,
        value=unicode_val)
    vuu__vgt = lir.FunctionType(lir.IntType(8).as_pointer(), [lir.IntType(8
        ).as_pointer(), lir.IntType(64)])
    qsa__scd = cgutils.get_or_insert_function(builder.module, vuu__vgt,
        name='init_string_const')
    return builder.call(qsa__scd, [tii__tqpg.data, tii__tqpg.length])


def gen_std_str_to_unicode(context, builder, std_str_val, del_str=False):
    kind = numba.cpython.unicode.PY_UNICODE_1BYTE_KIND

    def _std_str_to_unicode(std_str):
        length = bodo.libs.str_ext.get_std_str_len(std_str)
        pbymr__lhd = numba.cpython.unicode._empty_string(kind, length, 1)
        bodo.libs.str_arr_ext._memcpy(pbymr__lhd._data, bodo.libs.str_ext.
            get_c_str(std_str), length, 1)
        if del_str:
            bodo.libs.str_ext.del_str(std_str)
        return pbymr__lhd
    val = context.compile_internal(builder, _std_str_to_unicode,
        string_type(bodo.libs.str_ext.std_str_type), [std_str_val])
    return val


def gen_get_unicode_chars(context, builder, unicode_val):
    tii__tqpg = cgutils.create_struct_proxy(string_type)(context, builder,
        value=unicode_val)
    return tii__tqpg.data


@intrinsic
def unicode_to_std_str(typingctx, unicode_t=None):

    def codegen(context, builder, sig, args):
        return gen_unicode_to_std_str(context, builder, args[0])
    return std_str_type(string_type), codegen


@intrinsic
def std_str_to_unicode(typingctx, unicode_t=None):

    def codegen(context, builder, sig, args):
        return gen_std_str_to_unicode(context, builder, args[0], True)
    return string_type(std_str_type), codegen


class RandomAccessStringArrayType(types.ArrayCompatible):

    def __init__(self):
        super(RandomAccessStringArrayType, self).__init__(name=
            'RandomAccessStringArrayType()')

    @property
    def as_array(self):
        return types.Array(types.undefined, 1, 'C')

    @property
    def dtype(self):
        return string_type

    def copy(self):
        RandomAccessStringArrayType()


random_access_string_array = RandomAccessStringArrayType()


@register_model(RandomAccessStringArrayType)
class RandomAccessStringArrayModel(models.StructModel):

    def __init__(self, dmm, fe_type):
        gdmw__xtw = [('data', types.List(string_type))]
        models.StructModel.__init__(self, dmm, fe_type, gdmw__xtw)


make_attribute_wrapper(RandomAccessStringArrayType, 'data', '_data')


@intrinsic
def alloc_random_access_string_array(typingctx, n_t=None):

    def codegen(context, builder, sig, args):
        fura__zpr, = args
        kiys__cckf = types.List(string_type)
        bmvpl__ptrlk = numba.cpython.listobj.ListInstance.allocate(context,
            builder, kiys__cckf, fura__zpr)
        bmvpl__ptrlk.size = fura__zpr
        ogf__bms = cgutils.create_struct_proxy(sig.return_type)(context,
            builder)
        ogf__bms.data = bmvpl__ptrlk.value
        return ogf__bms._getvalue()
    return random_access_string_array(types.intp), codegen


@overload(operator.getitem, no_unliteral=True)
def random_access_str_arr_getitem(A, ind):
    if A != random_access_string_array:
        return
    if isinstance(ind, types.Integer):
        return lambda A, ind: A._data[ind]


@overload(operator.setitem)
def random_access_str_arr_setitem(A, idx, val):
    if A != random_access_string_array:
        return
    if isinstance(idx, types.Integer):
        assert val == string_type

        def impl_scalar(A, idx, val):
            A._data[idx] = val
        return impl_scalar


@overload(len, no_unliteral=True)
def overload_str_arr_len(A):
    if A == random_access_string_array:
        return lambda A: len(A._data)


@overload_attribute(RandomAccessStringArrayType, 'shape')
def overload_str_arr_shape(A):
    return lambda A: (len(A._data),)


def alloc_random_access_str_arr_equiv(self, scope, equiv_set, loc, args, kws):
    assert len(args) == 1 and not kws
    return ArrayAnalysis.AnalyzeResult(shape=args[0], pre=[])


(ArrayAnalysis.
    _analyze_op_call_bodo_libs_str_ext_alloc_random_access_string_array
    ) = alloc_random_access_str_arr_equiv
str_from_float32 = types.ExternalFunction('str_from_float32', types.void(
    types.voidptr, types.float32))
str_from_float64 = types.ExternalFunction('str_from_float64', types.void(
    types.voidptr, types.float64))


def float_to_str(s, v):
    pass


@overload(float_to_str)
def float_to_str_overload(s, v):
    assert isinstance(v, types.Float)
    if v == types.float32:
        return lambda s, v: str_from_float32(s._data, v)
    return lambda s, v: str_from_float64(s._data, v)


@overload(str)
def float_str_overload(v):
    if isinstance(v, types.Float):
        kind = numba.cpython.unicode.PY_UNICODE_1BYTE_KIND

        def impl(v):
            if v == 0:
                return '0.0'
            mbxio__hhzvs = 0
            qhhx__vefl = v
            if qhhx__vefl < 0:
                mbxio__hhzvs = 1
                qhhx__vefl = -qhhx__vefl
            if qhhx__vefl < 1:
                sumtr__drhvl = 1
            else:
                sumtr__drhvl = 1 + int(np.floor(np.log10(qhhx__vefl)))
            length = mbxio__hhzvs + sumtr__drhvl + 1 + 6
            s = numba.cpython.unicode._malloc_string(kind, 1, length, True)
            float_to_str(s, v)
            return s
        return impl


@overload(format, no_unliteral=True)
def overload_format(value, format_spec=''):
    if is_overload_constant_str(format_spec) and get_overload_const_str(
        format_spec) == '':

        def impl_fast(value, format_spec=''):
            return str(value)
        return impl_fast

    def impl(value, format_spec=''):
        with numba.objmode(res='string'):
            res = format(value, format_spec)
        return res
    return impl


@lower_cast(StdStringType, types.float64)
def cast_str_to_float64(context, builder, fromty, toty, val):
    vuu__vgt = lir.FunctionType(lir.DoubleType(), [lir.IntType(8).as_pointer()]
        )
    qsa__scd = cgutils.get_or_insert_function(builder.module, vuu__vgt,
        name='str_to_float64')
    res = builder.call(qsa__scd, (val,))
    bodo.utils.utils.inlined_check_and_propagate_cpp_exception(context, builder
        )
    return res


@lower_cast(StdStringType, types.float32)
def cast_str_to_float32(context, builder, fromty, toty, val):
    vuu__vgt = lir.FunctionType(lir.FloatType(), [lir.IntType(8).as_pointer()])
    qsa__scd = cgutils.get_or_insert_function(builder.module, vuu__vgt,
        name='str_to_float32')
    res = builder.call(qsa__scd, (val,))
    bodo.utils.utils.inlined_check_and_propagate_cpp_exception(context, builder
        )
    return res


@lower_cast(string_type, types.float64)
def cast_unicode_str_to_float64(context, builder, fromty, toty, val):
    std_str = gen_unicode_to_std_str(context, builder, val)
    return cast_str_to_float64(context, builder, std_str_type, toty, std_str)


@lower_cast(string_type, types.float32)
def cast_unicode_str_to_float32(context, builder, fromty, toty, val):
    std_str = gen_unicode_to_std_str(context, builder, val)
    return cast_str_to_float32(context, builder, std_str_type, toty, std_str)


@lower_cast(string_type, types.int64)
@lower_cast(string_type, types.int32)
@lower_cast(string_type, types.int16)
@lower_cast(string_type, types.int8)
def cast_unicode_str_to_int64(context, builder, fromty, toty, val):
    tii__tqpg = cgutils.create_struct_proxy(string_type)(context, builder,
        value=val)
    vuu__vgt = lir.FunctionType(lir.IntType(toty.bitwidth), [lir.IntType(8)
        .as_pointer(), lir.IntType(64)])
    qsa__scd = cgutils.get_or_insert_function(builder.module, vuu__vgt,
        name='str_to_int64')
    res = builder.call(qsa__scd, (tii__tqpg.data, tii__tqpg.length))
    bodo.utils.utils.inlined_check_and_propagate_cpp_exception(context, builder
        )
    return res


@lower_cast(string_type, types.uint64)
@lower_cast(string_type, types.uint32)
@lower_cast(string_type, types.uint16)
@lower_cast(string_type, types.uint8)
def cast_unicode_str_to_uint64(context, builder, fromty, toty, val):
    tii__tqpg = cgutils.create_struct_proxy(string_type)(context, builder,
        value=val)
    vuu__vgt = lir.FunctionType(lir.IntType(toty.bitwidth), [lir.IntType(8)
        .as_pointer(), lir.IntType(64)])
    qsa__scd = cgutils.get_or_insert_function(builder.module, vuu__vgt,
        name='str_to_uint64')
    res = builder.call(qsa__scd, (tii__tqpg.data, tii__tqpg.length))
    bodo.utils.utils.inlined_check_and_propagate_cpp_exception(context, builder
        )
    return res


@infer_getattr
class StringAttribute(AttributeTemplate):
    key = types.UnicodeType

    @bound_function('str.format', no_unliteral=True)
    def resolve_format(self, string_typ, args, kws):
        kws = dict(kws)
        mitm__lhd = ', '.join('e{}'.format(vkzn__ilhpo) for vkzn__ilhpo in
            range(len(args)))
        if mitm__lhd:
            mitm__lhd += ', '
        mfrno__bqqn = ', '.join("{} = ''".format(a) for a in kws.keys())
        yqvk__jed = f'def format_stub(string, {mitm__lhd} {mfrno__bqqn}):\n'
        yqvk__jed += '    pass\n'
        bcei__fwx = {}
        exec(yqvk__jed, {}, bcei__fwx)
        kqr__vvncw = bcei__fwx['format_stub']
        fqjtm__fmkl = numba.core.utils.pysignature(kqr__vvncw)
        jqqz__hos = (string_typ,) + args + tuple(kws.values())
        return signature(string_typ, jqqz__hos).replace(pysig=fqjtm__fmkl)


@numba.njit(cache=True)
def str_split(arr, pat, n):
    dczp__iuahd = pat is not None and len(pat) > 1
    if dczp__iuahd:
        aol__ugu = re.compile(pat)
        if n == -1:
            n = 0
    elif n == 0:
        n = -1
    bmvpl__ptrlk = len(arr)
    zer__ggx = 0
    kuq__adz = 0
    for vkzn__ilhpo in numba.parfors.parfor.internal_prange(bmvpl__ptrlk):
        if bodo.libs.array_kernels.isna(arr, vkzn__ilhpo):
            continue
        if dczp__iuahd:
            ivmy__meupr = aol__ugu.split(arr[vkzn__ilhpo], maxsplit=n)
        elif pat == '':
            ivmy__meupr = [''] + list(arr[vkzn__ilhpo]) + ['']
        else:
            ivmy__meupr = arr[vkzn__ilhpo].split(pat, n)
        zer__ggx += len(ivmy__meupr)
        for s in ivmy__meupr:
            kuq__adz += bodo.libs.str_arr_ext.get_utf8_size(s)
    epvec__nybht = bodo.libs.array_item_arr_ext.pre_alloc_array_item_array(
        bmvpl__ptrlk, (zer__ggx, kuq__adz), bodo.libs.str_arr_ext.
        string_array_type)
    qdg__vgt = bodo.libs.array_item_arr_ext.get_offsets(epvec__nybht)
    hhm__boe = bodo.libs.array_item_arr_ext.get_null_bitmap(epvec__nybht)
    psas__htdj = bodo.libs.array_item_arr_ext.get_data(epvec__nybht)
    mqkda__ydgqb = 0
    for dusgg__ger in numba.parfors.parfor.internal_prange(bmvpl__ptrlk):
        qdg__vgt[dusgg__ger] = mqkda__ydgqb
        if bodo.libs.array_kernels.isna(arr, dusgg__ger):
            bodo.libs.int_arr_ext.set_bit_to_arr(hhm__boe, dusgg__ger, 0)
            continue
        bodo.libs.int_arr_ext.set_bit_to_arr(hhm__boe, dusgg__ger, 1)
        if dczp__iuahd:
            ivmy__meupr = aol__ugu.split(arr[dusgg__ger], maxsplit=n)
        elif pat == '':
            ivmy__meupr = [''] + list(arr[dusgg__ger]) + ['']
        else:
            ivmy__meupr = arr[dusgg__ger].split(pat, n)
        rtl__ykwf = len(ivmy__meupr)
        for osw__exqq in range(rtl__ykwf):
            s = ivmy__meupr[osw__exqq]
            psas__htdj[mqkda__ydgqb] = s
            mqkda__ydgqb += 1
    qdg__vgt[bmvpl__ptrlk] = mqkda__ydgqb
    return epvec__nybht


@overload(hex)
def overload_hex(x):
    if isinstance(x, types.Integer):
        kind = numba.cpython.unicode.PY_UNICODE_1BYTE_KIND

        def impl(x):
            x = np.int64(x)
            if x < 0:
                sqyny__psavx = '-0x'
                x = x * -1
            else:
                sqyny__psavx = '0x'
            x = np.uint64(x)
            if x == 0:
                ufc__bzwuj = 1
            else:
                ufc__bzwuj = fast_ceil_log2(x + 1)
                ufc__bzwuj = (ufc__bzwuj + 3) // 4
            length = len(sqyny__psavx) + ufc__bzwuj
            output = numba.cpython.unicode._empty_string(kind, length, 1)
            bodo.libs.str_arr_ext._memcpy(output._data, sqyny__psavx._data,
                len(sqyny__psavx), 1)
            int_to_hex(output, ufc__bzwuj, len(sqyny__psavx), x)
            return output
        return impl


@register_jitable
def fast_ceil_log2(x):
    omsjs__edww = 0 if x & x - 1 == 0 else 1
    unme__nixz = [np.uint64(18446744069414584320), np.uint64(4294901760),
        np.uint64(65280), np.uint64(240), np.uint64(12), np.uint64(2)]
    rtdx__pogq = 32
    for vkzn__ilhpo in range(len(unme__nixz)):
        btu__bbdd = 0 if x & unme__nixz[vkzn__ilhpo] == 0 else rtdx__pogq
        omsjs__edww = omsjs__edww + btu__bbdd
        x = x >> btu__bbdd
        rtdx__pogq = rtdx__pogq >> 1
    return omsjs__edww


@intrinsic
def int_to_hex(typingctx, output, out_len, header_len, int_val):

    def codegen(context, builder, sig, args):
        output, out_len, header_len, int_val = args
        nquh__rmqt = cgutils.create_struct_proxy(sig.args[0])(context,
            builder, value=output)
        vuu__vgt = lir.FunctionType(lir.IntType(8).as_pointer(), [lir.
            IntType(8).as_pointer(), lir.IntType(64), lir.IntType(64)])
        jed__twqlw = cgutils.get_or_insert_function(builder.module,
            vuu__vgt, name='int_to_hex')
        ahvpk__zfxog = builder.inttoptr(builder.add(builder.ptrtoint(
            nquh__rmqt.data, lir.IntType(64)), header_len), lir.IntType(8).
            as_pointer())
        builder.call(jed__twqlw, (ahvpk__zfxog, out_len, int_val))
    return types.void(output, out_len, header_len, int_val), codegen


def alloc_empty_bytes_or_string_data(typ, kind, length, is_ascii=0):
    pass


@overload(alloc_empty_bytes_or_string_data)
def overload_alloc_empty_bytes_or_string_data(typ, kind, length, is_ascii=0):
    typ = typ.instance_type if isinstance(typ, types.TypeRef) else typ
    if typ == bodo.bytes_type:
        return lambda typ, kind, length, is_ascii=0: np.empty(length, np.uint8)
    if typ == string_type:
        return (lambda typ, kind, length, is_ascii=0: numba.cpython.unicode
            ._empty_string(kind, length, is_ascii))
    raise BodoError(
        f'Internal Error: Expected Bytes or String type, found {typ}')


def get_unicode_or_numpy_data(val):
    pass


@overload(get_unicode_or_numpy_data)
def overload_get_unicode_or_numpy_data(val):
    if val == string_type:
        return lambda val: val._data
    if isinstance(val, types.Array):
        return lambda val: val.ctypes
    raise BodoError(
        f'Internal Error: Expected String or Numpy Array, found {val}')
